import pytest

from harmonia.surveyor.synthesis import (
    generate_compression_matrix,
)


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_compression_matrix():
    print("Unimplemented test.")
    raise NotImplementedError
