"""
***************************************************************************
Algorithms (:mod:`~harmonia.algorithms`)
***************************************************************************

Provide algorithms for Fourier basis evaluation and integration, spectrum
discretisation and structured array manipulation for cosmological data.

"""
from .arrays import CartesianArray, SphericalArray
from .discretisation import DiscreteSpectrum
