import pytest

from harmonia.surveyor.definition import (
    generate_mask_by_sky_fraction,
    generate_mask_from_map,
    generate_selection_by_distribution,
    generate_selection_from_samples,
    generate_selection_samples,
)


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_mask_by_sky_fraction():
    print("Unimplemented test.")
    raise NotImplementedError


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_mask_from_map():
    print("Unimplemented test.")
    raise NotImplementedError


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_selection_by_distribution():
    print("Unimplemented test.")
    raise NotImplementedError


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_selection_from_samples():
    print("Unimplemented test.")
    raise NotImplementedError


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_selection_samples():
    print("Unimplemented test.")
    raise NotImplementedError
